def _drop_duplicate_positions(df, **kwargs):
    strand = kwargs.get("strand")
    keep = kwargs.get("keep")

    columns = ["Start", "End"]
    if strand:
        columns.append("Strand")

    return df.drop_duplicates(columns, keep=keep)



import pandas as pd
import numpy as np


def _subseq(scdf, **kwargs):
    if scdf.empty:
        return None

    scdf = scdf.copy()
    orig_order = scdf.index.copy()
    scdf.insert(0, "__i__", orig_order)

    by = kwargs.get("by") if kwargs.get("by") else "__i__"
    if not type(by) is list:
        by = [by]

    strand = kwargs.get("strand")
    # at this point, strand is False if:
    #   1. subsequence was called with strand=False or
    #   2. it was called with strand=None and self is not stranded
    # or it can be '+' or '-' if:
    #   1. it was input as True to subsequence and passed  to pyrange_apply_single as True,
    #      which updates it to '-' or '+' before calling _subseq, or
    #   2. it was called with strand=None and self is stranded
    if strand != "-":
        strand = "+"
    # now, unstranded or strand==None cases are treated like all intervals are on the + strand

    # creating j which holds the boundaries per group
    # j contains one row per group; columns: Start  End (+ by columns); indexed by __i__
    agg_dict = {"__i__": "first", "Start": "min", "End": "max"}
    for b in by:
        agg_dict[b] = "first"

    if kwargs.get("by"):
        j = (
            scdf.groupby(by, dropna=False)[["Start", "End", "__i__"] + by]
            .agg(agg_dict)
            .set_index("__i__")
        )
    else:
        j = (
            scdf.groupby(by, dropna=False)[["Start", "End", "__i__"]]
            .agg(agg_dict)
            .set_index("__i__")
        )
        j.insert(0, "__i__", j.index)
        j.index.name = None

    start = kwargs.get("start") if kwargs.get("start") else 0
    end = kwargs.get("end") if kwargs.get("end") else (j.End - j.Start).max()

    # below we add columns starts, ends to j
    # start and ends define the desired start and end, with one row per group (transcript).
    # they may be out of bounds of the interval, though (this is dealt with later)

    if (strand == "+" and start >= 0) or (strand == "-" and start < 0):
        j["starts"] = j.Start + abs(start)
    else:
        j["starts"] = j.End - abs(start)

    if (strand == "+" and end >= 0) or (strand == "-" and end < 0):
        j["ends"] = j.Start + abs(end)
    else:
        j["ends"] = j.End - abs(end)

    if strand == "-":
        j.rename(columns={"ends": "__min__", "starts": "__max__"}, inplace=True)
    else:
        j.rename(columns={"starts": "__min__", "ends": "__max__"}, inplace=True)

    # I'm maintaing the original row order
    scdf = (
        scdf.merge(j[by + ["__min__", "__max__"]], on=by)
        .set_index("__i__")
        .loc[orig_order]
    )

    # instead of simply using starts and ends as computed above, we're dealing here with potential out of bounds:
    r = scdf[~((scdf.Start >= scdf.__max__) | (scdf.End <= scdf.__min__))].copy()
    r.loc[:, "Start"] = np.maximum(r.Start, r.__min__)
    r.loc[:, "End"] = np.minimum(r.End, r.__max__)

    r = r.drop(["__min__", "__max__"], axis=1)

    return r



import pandas as pd
import numpy as np

from pyranges.subset import get_string, get_tuple, get_slice, get_booldict
from pyranges.methods.drop import _keep

from pyranges import PyRanges


def _getitem(self, val):
    if isinstance(val, list):
        dfs = _keep(self, keep=val).dfs
    elif isinstance(val, str):
        dfs = get_string(self, val)
    elif isinstance(val, tuple):
        dfs = get_tuple(self, val)
    elif isinstance(val, slice):
        dfs = get_slice(self, val)
    elif isinstance(val, dict):
        dfs = get_booldict(self, val)
    elif (isinstance(val, (pd.Series, np.ndarray))) and val.dtype == "bool":
        assert len(val) == len(self), "Boolean indexer must be same length as pyrange!"
        _length = 0
        if isinstance(val, pd.Series):
            val = val.values

        dfs = {}
        for k, df in self:
            length = len(df)
            _bool = val[_length : (length + _length)]
            dfs[k] = df[_bool]
            _length += length
    else:
        raise Exception("Not a valid subsetter: {}".format(str(val)))

    gr = PyRanges(dfs)
    return gr



import sys
import pyranges as pr
import pandas as pd


def get_sequence(gr, path=None, pyfaidx_fasta=None):
    """Get the sequence of the intervals from a fasta file

    Parameters
    ----------
    gr : PyRanges

        Coordinates.

    path : str

        Path to fasta file. It will be indexed using pyfaidx if an index is not found

    pyfaidx_fasta : pyfaidx.Fasta

        Alternative method to provide fasta target, as a pyfaidx.Fasta object


    Returns
    -------
    Series

        Sequences, one per interval.

    Note
    ----

    This function requires the library pyfaidx, it can be installed with
    ``conda install -c bioconda pyfaidx`` or ``pip install pyfaidx``.

    Sorting the PyRanges is likely to improve the speed.
    Intervals on the negative strand will be reverse complemented.

    Warning
    -------

    Note that the names in the fasta header and gr must be the same.

    See also
    --------
    get_transcript_sequence : obtain mRNA sequences, by joining exons belonging to the same transcript


    Examples
    --------

    >>> gr = pr.from_dict({"Chromosome": ["chr1", "chr1"],
    ...                    "Start": [5, 0], "End": [8, 5]})

    >>> gr
    +--------------+-----------+-----------+
    | Chromosome   |     Start |       End |
    | (category)   |   (int32) |   (int32) |
    |--------------+-----------+-----------|
    | chr1         |         5 |         8 |
    | chr1         |         0 |         5 |
    +--------------+-----------+-----------+
    Unstranded PyRanges object has 2 rows and 3 columns from 1 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    >>> tmp_handle = open("temp.fasta", "w+")
    >>> _ = tmp_handle.write(">chr1\\n")
    >>> _ = tmp_handle.write("ATTACCAT\\n")
    >>> tmp_handle.close()

    >>> seq = pr.get_sequence(gr, "temp.fasta")

    >>> seq
    0      CAT
    1    ATTAC
    dtype: object

    >>> gr.seq = seq
    >>> gr
    +--------------+-----------+-----------+------------+
    | Chromosome   |     Start |       End | seq        |
    | (category)   |   (int32) |   (int32) | (object)   |
    |--------------+-----------+-----------+------------|
    | chr1         |         5 |         8 | CAT        |
    | chr1         |         0 |         5 | ATTAC      |
    +--------------+-----------+-----------+------------+
    Unstranded PyRanges object has 2 rows and 4 columns from 1 chromosomes.
    For printing, the PyRanges was sorted on Chromosome.

    """

    try:
        import pyfaidx
    except ModuleNotFoundError as e:
        print(
            "pyfaidx must be installed to get fasta sequences. Use `conda install -c bioconda pyfaidx` or `pip install pyfaidx` to install it."
        )
        sys.exit(1)

    if pyfaidx_fasta is None:
        if path is None:
            raise Exception(
                "ERROR get_sequence : you must provide a fasta path or pyfaidx_fasta object"
            )
        pyfaidx_fasta = pyfaidx.Fasta(path, read_ahead=int(1e5))

    seqs = []
    for k, df in gr:
        if type(k) is tuple:  # input is Stranded
            _fasta = pyfaidx_fasta[k[0]]
            if k[1] == "-":
                for start, end in zip(df.Start, df.End):
                    seqs.append((-_fasta[start:end]).seq)  # reverse complement
            else:
                for start, end in zip(df.Start, df.End):
                    seqs.append(_fasta[start:end].seq)

        else:
            _fasta = pyfaidx_fasta[k]
            for start, end in zip(df.Start, df.End):
                seqs.append(_fasta[start:end].seq)

    return pd.concat([pd.Series(s) for s in seqs]).reset_index(drop=True)


def get_fasta(*args, **kwargs):
    """Deprecated: this function has been moved to Pyranges.get_sequence"""
    return get_sequence(*args, **kwargs)


def get_transcript_sequence(gr, group_by, path=None, pyfaidx_fasta=None):
    """Get the sequence of mRNAs, e.g. joining intervals corresponding to exons of the same transcript

    Parameters
    ----------
    gr : PyRanges

        Coordinates.

    group_by : str or list of str

        intervals are grouped by this/these ID column(s): these are exons belonging to same transcript

    path : str

        Path to fasta file. It will be indexed using pyfaidx if an index is not found

    pyfaidx_fasta : pyfaidx.Fasta

        Alternative method to provide fasta target, as a pyfaidx.Fasta object


    Returns
    -------
    DataFrame

        Pandas DataFrame with a column for Sequence, plus ID column(s) provided with "group_by"

    Note
    ----

    This function requires the library pyfaidx, it can be installed with
    ``conda install -c bioconda pyfaidx`` or ``pip install pyfaidx``.

    Sorting the PyRanges is likely to improve the speed.
    Intervals on the negative strand will be reverse complemented.

    Warning
    -------

    Note that the names in the fasta header and gr must be the same.

    See also
    --------
    get_sequence : obtain sequence of single intervals


    Examples
    --------

    >>> gr = pr.from_dict({"Chromosome": ['chr1', 'chr1', 'chr1'],
    ...                    "Start": [0, 9, 18], "End": [4, 13, 21],
    ...                    "Strand":['+', '-', '-'],
    ...                    "transcript": ['t1', 't2', 't2']})
    >>> gr
    +--------------+-----------+-----------+--------------+--------------+
    | Chromosome   |     Start |       End | Strand       | transcript   |
    | (category)   |   (int32) |   (int32) | (category)   | (object)     |
    |--------------+-----------+-----------+--------------+--------------|
    | chr1         |         0 |         4 | +            | t1           |
    | chr1         |         9 |        13 | -            | t2           |
    | chr1         |        18 |        21 | -            | t2           |
    +--------------+-----------+-----------+--------------+--------------+
    Stranded PyRanges object has 3 rows and 5 columns from 1 chromosomes.
    For printing, the PyRanges was sorted on Chromosome and Strand.


    >>> tmp_handle = open("temp.fasta", "w+")
    >>> _ = tmp_handle.write(">chr1\\n")
    >>> _ = tmp_handle.write("AAACCCTTTGGGAAACCCTTTGGG\\n")
    >>> tmp_handle.close()

    >>> seq = pr.get_transcript_sequence(gr, path="temp.fasta", group_by='transcript')
    >>> seq
      transcript Sequence
    0         t1     AAAC
    1         t2  AAATCCC

    To write to a file in fasta format:
    # with open('outfile.fasta', 'w') as fw:
    #     nchars=60
    #     for row in seq.itertuples():
    #         s = '\\n'.join([ row.Sequence[i:i+nchars] for i in range(0, len(row.Sequence), nchars)])
    #         fw.write(f'>{row.transcript}\\n{s}\\n')
    """

    if gr.stranded:
        gr = gr.sort("5")
    else:
        gr = gr.sort()

    z = gr.df
    z["Sequence"] = get_sequence(gr, path=path, pyfaidx_fasta=pyfaidx_fasta)

    return z.groupby(group_by, as_index=False).agg({"Sequence": "".join})



import pandas as pd


def _sort(df, **kwargs):
    if "by" in kwargs:
        by = kwargs["by"]
        if type(kwargs["by"]) is str:
            by = [by]
        else:
            by = [f for f in by]  # making sure it's a list

        if "5" in by:
            i = by.index("5")
            byp = by.copy()
            byp[i : i + 1] = ["Start", "End"]
            byn = by.copy()
            byn[i : i + 1] = ["End", "Start"]
            byna = [False if j in (i, i + 1) else True for j, f in enumerate(byn)]

            return pd.concat(
                [
                    df[df.Strand == "+"].sort_values(byp, ascending=True),
                    df[df.Strand == "-"].sort_values(byn, ascending=byna),
                ]
            )

        else:
            by = kwargs["by"]
            return df.sort_values(by)
    else:
        df = sort_one_by_one(df, "Start", "End")
        return df


def sort_one_by_one(d, col1, col2):
    """
    Equivalent to pd.sort_values(by=[col1, col2]), but faster.
    """

    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind="mergesort")



import pandas as pd
import pyranges as pr


def _handle_eval_return(self, result, col, as_pyranges, subset):
    """Handle return from eval.

    If col is set, add/update cols. If subset is True, use return series to subset PyRanges.
    Otherwise return PyRanges or dict of data."""

    if as_pyranges:
        if not result:
            return pr.PyRanges()

        first_hit = list(result.values())[0]

        if isinstance(first_hit, pd.Series):
            if first_hit.dtype == bool and subset:
                return self[result]
            elif col:
                self.__setattr__(col, result)
                return self
            else:
                raise Exception(
                    "Cannot return PyRanges when function returns a Series! Use as_pyranges=False."
                )
        return pr.PyRanges(result)
    else:
        return result


def _call(self, f, strand=None, as_pyranges=True, **kwargs):
    if strand is None:
        strand = self.stranded

    if self.stranded and not strand:
        self = self.unstrand()

    result = self.apply(f, strand=strand, as_pyranges=False, **kwargs)

    # result = _handle_eval_return(self, result, col, as_pyranges, subset)

    return result



import pandas as pd
import numpy as np


def _spliced_subseq(scdf, **kwargs):
    if scdf.empty:
        return None

    scdf = scdf.copy()
    orig_order = scdf.index.copy()

    by = kwargs.get("by") if kwargs.get("by") else "__i__"
    if not type(by) is list:
        by = [by]

    strand = kwargs.get("strand")

    # at this point, strand is False if 1. spliced_subsequence was called with strand=False or
    #                                   2. it was called with strand=None and self is not stranded
    # or it can be '+' or '-' if:
    #  1. it was input as True to spliced_subsequence and passed  to pyrange_apply_single as True,
    #     which updates it to '-' or '+' before calling _spliced_subseq, or
    #  2. it was called with strand=None and self is stranded

    if strand:
        assert "Strand" in scdf, "Cannot have strand=True on unstranded pyranges!"

    scdf.insert(scdf.shape[1], "__length__", scdf.End - scdf.Start)
    scdf.insert(scdf.shape[1], "__i__", scdf.index)

    g = scdf.groupby(by, dropna=False)
    scdf.insert(scdf.shape[1], "__cumsum__", g.__length__.cumsum())

    start = kwargs.get("start") if kwargs.get("start") else 0
    end = kwargs.get("end") if kwargs.get("end") else scdf.__cumsum__.max()

    minstart_idx = g.__i__.first()

    if start < 0 or (end is not None and end < 0):
        # len_per_transc is total sum of exon length per transcript
        len_per_transc = scdf.loc[g.__i__.last(), by + ["__cumsum__"]].rename(
            columns={"__cumsum__": "__totexonlen__"}
        )

        # exp_len_per_transc has same rows of scdf with total sum of exon length
        # had to add bits to keep the order of rows right, or merge would destroy it
        if kwargs.get("by"):
            exp_len_per_transc = (
                scdf.loc[:, by + ["__i__"]]
                .merge(len_per_transc, on=by)
                .set_index("__i__")
                .loc[scdf.index]
            )
        else:
            exp_len_per_transc = (
                scdf.loc[:, by]
                .merge(len_per_transc, on=by)
                .set_index("__i__")
                .loc[scdf.index]
            )

        if start < 0:
            start = exp_len_per_transc["__totexonlen__"] + start

        if end is not None and end < 0:
            end = exp_len_per_transc["__totexonlen__"] + end

    cs_start = g.__cumsum__.shift(1, fill_value=0)
    cs_start.loc[minstart_idx] = 0

    cs_end = scdf["__cumsum__"]

    ## NOTE
    #  here below, start is a scalar if originally provided > 0, or a Series if < 0
    #              end is a scalar if originally None or provided >0, or a Series if provided < 0
    if strand == "-":  # and use_strand:
        start_adjustments = start - cs_start
        adjust_start = start_adjustments > 0
        scdf.loc[adjust_start, "End"] -= start_adjustments[adjust_start].astype(
            scdf.End.dtype
        )

        end_adjustments = cs_end - end
        adjust_end = end_adjustments > 0
        scdf.loc[adjust_end, "Start"] += end_adjustments[adjust_end].astype(
            scdf.Start.dtype
        )
    else:
        start_adjustments = start - cs_start
        adjust_start = start_adjustments > 0
        scdf.loc[adjust_start, "Start"] += start_adjustments[adjust_start].astype(
            scdf.Start.dtype
        )

        end_adjustments = cs_end - end
        adjust_end = end_adjustments > 0
        scdf.loc[adjust_end, "End"] -= end_adjustments[adjust_end].astype(
            scdf.End.dtype
        )

    scdf = scdf.loc[orig_order]
    scdf = scdf[(scdf.Start < scdf.End)]

    return scdf.drop(["__i__", "__length__", "__cumsum__"], axis=1)



import pandas as pd


from .join import _both_dfs
from .sort import sort_one_by_one

from sorted_nearest import (
    nearest_previous_nonoverlapping,
    nearest_next_nonoverlapping,
    nearest_nonoverlapping,
)


def _insert_distance(ocdf, dist, suffix):
    if "Distance" not in ocdf:
        distance_column_name = "Distance"
    elif "Distance" + suffix not in ocdf:
        distance_column_name = "Distance" + suffix
    else:
        i = 1
        while "Distance" + str(i) in ocdf:
            i += 1
        distance_column_name = "Distance" + str(i)

    ocdf.insert(
        ocdf.shape[1],
        distance_column_name,
        pd.Series(dist, index=ocdf.index).fillna(-1).astype(int),
    )

    return ocdf


def _overlapping_for_nearest(scdf, ocdf, suffix):
    nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    scdf2, ocdf2 = _both_dfs(scdf, ocdf, how="first")

    if not ocdf2.empty:
        original_idx = scdf.index

        idxs = scdf2.index
        original_idx = scdf.index.copy(deep=True)
        missing_idxs = ~original_idx.isin(idxs)
        missing_overlap = scdf.index[missing_idxs]

        df_to_find_nearest_in = scdf.reindex(missing_overlap)

        odf = ocdf.reindex(ocdf2.index)
        odf.index = idxs
        sdf = scdf.reindex(idxs)

        nearest_df = sdf.join(odf, rsuffix=suffix)
        nearest_df = _insert_distance(nearest_df, 0, suffix)
    else:
        df_to_find_nearest_in = scdf

    return nearest_df, df_to_find_nearest_in


def _next_nonoverlapping(left_ends, right_starts, right_indexes):
    left_ends = left_ends.sort_values()
    right_starts = right_starts.sort_values()
    r_idx, dist = nearest_next_nonoverlapping(
        left_ends.values - 1, right_starts.values, right_indexes
    )
    r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
    dist = pd.Series(dist, index=left_ends.index).sort_index().values

    return r_idx, dist


def _previous_nonoverlapping(left_starts, right_ends):
    left_starts = left_starts.sort_values()
    right_ends = right_ends.sort_values()
    r_idx, dist = nearest_previous_nonoverlapping(
        left_starts.values, right_ends.values - 1, right_ends.index.values
    )

    r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
    dist = pd.Series(dist, index=left_starts.index).sort_index().values

    return r_idx, dist


def _nearest(scdf, ocdf, **kwargs):
    if scdf.empty or ocdf.empty:
        return None

    overlap = kwargs["overlap"]
    how = kwargs["how"]
    suffix = kwargs["suffix"]

    if how == "upstream":
        strand = scdf.Strand.iloc[0]
        how = {"+": "previous", "-": "next"}[strand]
    elif how == "downstream":
        strand = scdf.Strand.iloc[0]
        how = {"+": "next", "-": "previous"}[strand]

    ocdf = ocdf.reset_index(drop=True)

    if overlap:
        nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(scdf, ocdf, suffix)
    else:
        df_to_find_nearest_in = scdf

    if not df_to_find_nearest_in.empty:
        df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
        ocdf = sort_one_by_one(ocdf, "Start", "End")
        df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

        if how == "next":
            r_idx, dist = _next_nonoverlapping(
                df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values
            )
        elif how == "previous":
            r_idx, dist = _previous_nonoverlapping(
                df_to_find_nearest_in.Start, ocdf.End
            )
        else:
            previous_r_idx, previous_dist = _previous_nonoverlapping(
                df_to_find_nearest_in.Start, ocdf.End
            )

            next_r_idx, next_dist = _next_nonoverlapping(
                df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values
            )

            r_idx, dist = nearest_nonoverlapping(
                previous_r_idx, previous_dist, next_r_idx, next_dist
            )

        ocdf = ocdf.reindex(r_idx)

        ocdf.index = df_to_find_nearest_in.index

        ocdf = _insert_distance(ocdf, dist, suffix)

        r_idx = pd.Series(r_idx, index=ocdf.index)
        df_to_find_nearest_in = df_to_find_nearest_in.drop(r_idx.loc[r_idx == -1].index)

        df = df_to_find_nearest_in.join(ocdf, rsuffix=suffix)

    if overlap and "df" in locals() and not df.empty and not nearest_df.empty:
        df = pd.concat([nearest_df, df], sort=False)
    elif overlap and not nearest_df.empty:
        df = nearest_df

    df = df.drop("Chromosome" + suffix, axis=1)
    return df



import numpy as np
import pandas as pd
from ncls import NCLS


def _intersection(scdf, ocdf, **kwargs):
    how = kwargs["how"]

    if ocdf.empty or scdf.empty:
        return None

    assert how in "containment first last".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    in_dtype = ocdf.Start.dtype

    oncls = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how or how is None:
        _self_indexes, _other_indexes = oncls.all_overlaps_both(starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = oncls.all_containments_both(
            starts, ends, indexes
        )
    elif how == "first":
        _self_indexes, _other_indexes = oncls.first_overlap_both(starts, ends, indexes)
    elif how == "last":
        _self_indexes, _other_indexes = oncls.last_overlap_both(starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    scdf, ocdf = scdf.reindex(_self_indexes), ocdf.reindex(_other_indexes)

    new_starts = pd.Series(
        np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
        index=scdf.index,
        dtype=in_dtype,
    )

    new_ends = pd.Series(
        np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
        index=scdf.index,
        dtype=in_dtype,
    )

    pd.options.mode.chained_assignment = None  # default='warn'
    scdf.loc[:, "Start"] = new_starts
    scdf.loc[:, "End"] = new_ends
    pd.options.mode.chained_assignment = "warn"

    if not scdf.empty:
        return scdf
    else:
        return None


def _overlap(scdf, ocdf, **kwargs):
    return_indexes = kwargs.get("return_indexes", False)

    if scdf.empty or ocdf.empty:
        return None

    how = kwargs["how"]

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _indexes = it.all_overlaps_self(starts, ends, indexes)
    elif how == "containment":
        _indexes, _ = it.all_containments_both(starts, ends, indexes)
    else:
        _indexes = it.has_overlaps(starts, ends, indexes)

    if return_indexes:
        return _indexes

    return scdf.reindex(_indexes)


def _count_overlaps(scdf, ocdf, **kwargs):
    kwargs["return_indexes"] = True
    idx = _overlap(scdf, ocdf, **kwargs)

    sx = pd.DataFrame(np.zeros(len(scdf), dtype=np.int32), index=scdf.index)

    vc = pd.Series(idx, dtype=np.int64).value_counts(sort=False)

    sx.loc[vc.index, 0] = vc.values

    scdf.insert(scdf.shape[1], kwargs["name"], sx)

    return scdf


# def _first_df(scdf, ocdf, kwargs):

#     if scdf.empty or ocdf.empty:
#         return None

#     how = kwargs["how"]

#     assert how in "containment first".split() + [False, None]
#     starts = scdf.Start.values
#     ends = scdf.End.values
#     indexes = scdf.index.values

#     it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

#     if not how:
#         _indexes = it.has_overlaps(starts, ends, indexes)
#     elif how == "containment":
#         _indexes = it.has_containment(starts, ends, indexes)

#     return scdf.reindex(_indexes)



# from pyranges.methods.join import _both_dfs

# import numpy as np
# import pandas as pd

# def _insert(scdf, ocdf, kwargs):

#     how = kwargs.get("how")
#     col = kwargs.get("columns")
#     suffix = kwargs.get("suffix")
#     overlap_only = kwargs.get("overlap_only")

#     if scdf.empty or ocdf.empty:
#         return None

#     _scdf, ocdf = _both_dfs(scdf, ocdf, how=how)

#     nix = pd.Index(range(len(_scdf)))
#     _scdf.index = nix
#     ocdf.index = nix

#     ocdf = ocdf[col]

#     df = _scdf.join(ocdf, rsuffix=suffix)

#     if not overlap_only:
#         # need to insert nan for those rows in self which had no overlap

#         scdf = scdf[~scdf.index.isin(_scdf.index)]

#         if isinstance(col, str):
#             scdf.insert(scdf.shape[1], col, np.nan)
#         else:
#             for c in col:
#                 scdf.insert(scdf.shape[1], c, np.nan)

#         df = pd.concat([df, scdf])

#     df = df.sort_values("Start End".split())

#     return df



def _bounds(scdf, **kwargs):
    if scdf.empty:
        return None

    col_order = [c for c in scdf.columns]

    by = kwargs.get("group_by")
    if not type(by) is list:
        by = [by]

    agg_dict = kwargs.get("agg") if kwargs.get("agg") else {}
    agg_dict.update({"Start": "min", "End": "max", "Chromosome": "first"})
    if "Strand" in scdf.columns:
        agg_dict["Strand"] = "first"

    res = scdf.groupby(by).agg(agg_dict).reset_index()
    res = res.reindex(columns=[c for c in col_order if c in res.columns])

    return res



from sorted_nearest import makewindows
from sorted_nearest import maketiles

import numpy as np


def _windows(df, **kwargs):
    window_size = kwargs["window_size"]

    idxs, starts, ends = makewindows(
        df.index.values, df.Start.values, df.End.values, window_size
    )

    df = df.reindex(idxs)
    df.loc[:, "Start"] = starts
    df.loc[:, "End"] = ends

    return df


def _intersect_tile(df):
    overlap = np.minimum(df.End, df.__End__) - np.maximum(df.Start, df.__Start__)
    df.insert(df.shape[1], "TileOverlap", overlap)

    return df


def _tiles(df, **kwargs):
    overlap = kwargs.get("overlap")

    if overlap:
        df = df.copy()
        df.insert(df.shape[1], "__Start__", df.Start)
        df.insert(df.shape[1], "__End__", df.End)

    window_size = kwargs["tile_size"]

    idxs, starts, ends = maketiles(
        df.index.values, df.Start.values, df.End.values, window_size
    )

    df = df.reindex(idxs)
    df.loc[:, "Start"] = starts
    df.loc[:, "End"] = ends

    if overlap:
        df = _intersect_tile(df)

        df = df.drop(["__Start__", "__End__"], axis=1)

    return df



from ncls import NCLS
import pandas as pd


def create_ncls(df):
    return NCLS(df.Start.values, df.End.values, df.index.values)


def find_overlaps(df, start, end):
    n = create_ncls(df)

    idxes = []
    for r in n.find_overlap(start, end):
        idxes.append(r[2])

    return idxes


def get_slice(self, val):
    # 100:999

    d = {}

    for k, df in self.items():
        start = val.start or 0
        stop = val.stop or max(df.End.max(), start)
        idxes = find_overlaps(df, start, stop)
        d[k] = df.reindex(idxes)

    return d


def get_string(self, val):
    if val in self.chromosomes:
        if self.stranded:
            return {k: self.dfs[k] for k in self.keys() if k[0] == val}
        else:
            return {val: self.dfs[val]}

    elif val in "+ -".split():
        return {k: v for k, v in self.items() if k[1] == val}
    else:
        return {}


def get_tuple(self, val):
    if len(val) == 2:
        dfs = get_double(self, val)
    elif len(val) == 3:
        dfs = get_triple(self, val)

    return dfs


def get_double(self, val):
    if len(val) == 2 and val[0] in self.chromosomes and isinstance(val[1], slice):
        chromosome, loc = val
        start = loc.start or 0
        if self.stranded:
            dfs = {k: df for k, df in self.items() if k[0] == chromosome}
            max_end = max([df.End.max() for df in dfs.values()])
        else:
            dfs = {val[0]: self.dfs[val[0]]}
            max_end = list(dfs.values())[0].End.max()

        # in case 1:None
        stop = loc.stop or max(max_end, start)

        dfs2 = {}
        for k, df in dfs.items():
            idxes = find_overlaps(df, start, stop)
            if idxes:
                dfs2[k] = df.loc[idxes]

        return dfs2

    # "+", 5:10
    if len(val) == 2 and val[0] in "+ -".split() and isinstance(val[1], slice):
        strand, loc = val
        start = loc.start or 0

        dfs = {k: df for k, df in self.items() if k[1] == strand}
        max_end = max([df.End.max() for df in dfs.values()])

        stop = loc.stop or max(max_end, start)

        dfs2 = {}
        for k, df in dfs.items():
            idxes = find_overlaps(df, start, stop)
            if idxes:
                dfs2[k] = df.loc[idxes]

        return dfs2

    # "chr1", "+"
    if len(val) == 2 and val[1] in "+ -".split():
        chromosome, strand = val

        if (chromosome, strand) in self.dfs:
            return {(chromosome, strand): self.dfs[chromosome, strand]}
        else:
            return {}


def get_triple(self, val):
    # "chr1", "+", 5:10
    chromosome, strand, loc = val
    start = loc.start or 0

    if strand not in "+ -".split():
        raise Exception("Strand '{}' invalid.".format(val))

    r = self[chromosome, strand].values()
    if len(r):
        df = r[0]
    else:
        df = pd.DataFrame(columns="Chromosome Start End".split())
        return df

    max_end = df.End.max()

    stop = loc.stop or max(max_end, start)

    idxes = find_overlaps(df, start, stop)
    return {(chromosome, strand): df.reindex(idxes)}


def get_booldict(self, df):
    _overlapping = set(self.dfs.keys()).intersection(set(df.keys()))

    new_dfs = {}
    for k in _overlapping:
        new_dfs[k] = self.dfs[k][df[k]]

    return new_dfs



import pyranges as pr
import numpy as np

np.random.seed(0)


def test_stranded():
    cpg = pr.data.cpg()
    exons = pr.data.exons()

    j = cpg.join(exons)

    assert j.stranded

    j.Strand = "."

    assert not j.stranded

    j.Strand = np.random.choice("+ -".split(), size=len(j))

    assert j.stranded

    for _, df in j:
        assert len(df.Strand.drop_duplicates()) == 1


def test_unstrand():
    exons = pr.data.exons()

    cpg = pr.data.cpg()
    print(exons)
    print(exons.columns)
    x = exons.unstrand()
    print(x.stranded)
    print(x)
    print(x.columns)
    for _, df in x:
        assert not df.index.duplicated().sum()

    # x = pr.PyRanges(x.df.reset_index(drop=True))
    cpg.join(x)



import numpy as np


def _relative_distance(scdf, ocdf, **kwargs):
    if scdf.empty or ocdf.empty:
        return np.array([])

    midpoints_self = ((scdf.Start + scdf.End) / 2).astype(int).sort_values().values
    midpoints_other = ((ocdf.Start + ocdf.End) / 2).astype(int).sort_values().values

    left_idx = np.searchsorted(midpoints_other, midpoints_self)
    left_idx[left_idx >= len(midpoints_other)] -= 1
    left_idx_shift = (midpoints_other[left_idx] > midpoints_self) & (left_idx > 0)
    left_idx[left_idx_shift] -= 1
    left = midpoints_other[left_idx]
    right_idx = np.searchsorted(midpoints_other, midpoints_self, side="right")
    right_idx[right_idx >= len(midpoints_other)] -= 1

    if len(right_idx) == 0 or len(left_idx) == 0:
        return np.array([])

    right = midpoints_other[right_idx]

    left_distance = np.absolute(midpoints_self - left)
    right_distance = np.absolute(midpoints_self - right)

    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.minimum(left_distance, right_distance) / (right - left)

    result[np.isnan(result)] = 0

    result = result[~np.isinf(result)]

    return result



import pytest

import pyranges as pr
import pandas as pd

from io import StringIO


@pytest.fixture
def expected_result():
    c = """Chromosome  Start  End       Name  Score Strand
0       chr1      0   10  interval1      0      +
1       chr1      0   10  interval3      0      +
2       chr1      0   10  interval2      0      -"""

    df = pd.read_csv(StringIO(c), sep=r"\s+")

    return df


def test_windows():
    f1 = pr.data.f1()

    print(f1)

    result = f1.tile(2)

    print(result)

    df = result.df

    assert list(df.Start) == [2, 4, 8, 4, 6]
    assert list(df.End) == [4, 6, 10, 6, 8]


def test_windows2():
    c = """Chromosome  Start    End  Count
0       chr1  10200  10400      7
1       chr1  10400  10600      7
2       chr1  51400  51600      1
3       chr1  51600  51800      3
4       chr1  51800  52000      3"""

    df = pd.read_csv(StringIO(c), sep=r"\s+", nrows=5)
    # df.End -= 1
    gr = pr.PyRanges(df)

    print(gr)
    result = gr.tile(200)

    print(result)

    df = result.df

    assert list(df.Start) == [10200, 10400, 51400, 51600, 51800]
    # assert list(df.End) == [4, 6, 8, 10, 6, 8]



#!/usr/bin/env python

from sorted_nearest import max_disjoint


def _max_disjoint(df, **kwargs):
    if df.empty:
        return None

    slack = kwargs.get("slack", 0)

    cdf = df.sort_values("End")

    idx = max_disjoint(cdf.index.values, cdf.Start.values, cdf.End.values, slack)

    return cdf.reindex(idx)



import pyranges as pr


def test_read_gtf():
    gr = pr.read_gtf("tests/test_data/ensembl.gtf", full=True)

    assert len(gr.columns) == 26

    df = gr.df
    transcript = df.iloc[1]
    assert transcript["tag"] == "basic"

    exon = df[df["exon_id"] == "ENSE00003812156"].iloc[0]
    assert exon["tag"] == "basic"

    gr = pr.read_gtf("tests/test_data/ensembl.gtf", full=True, duplicate_attr=True)
    print(gr.columns)
    assert len(gr.columns) == 26

    df = gr.df
    transcript = df.iloc[1]
    assert transcript["tag"] == "basic"

    exon = df[df["exon_id"] == "ENSE00003812156"].iloc[0]
    assert exon["tag"] == "CCDS,basic"
    # assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()


def test_read_gff3():
    gr = pr.read_gff3("tests/test_data/gencode.gff3")

    assert len(gr.columns) == 26
    # assert list(gr.df.columns[:4]) == "Chromosome Start End Strand".split()


def test_read_bed():
    pr.read_bed("pyranges/example_data/chipseq.bed")



from collections.abc import Iterable


def _drop(self, drop=None, like=None):
    columns = self.columns

    if isinstance(drop, str):
        drop = [drop]

    if not drop is None:
        for i in "Chromosome Start End".split():
            assert not i in drop, "Cannot drop {}".format(i)

    want_to_drop_strand = (
        isinstance(drop, str)
        and drop == "Strand"
        or (isinstance(drop, list) and "Strand" in drop)
    )
    if not self.stranded or want_to_drop_strand:
        always_keep = "Chromosome Start End".split()
    else:
        always_keep = "Chromosome Start End Strand".split()

    _to_drop = []

    if like:
        import re

        r = re.compile(like)
        _to_drop = [c for c in self.columns if r.search(c) is not None]
    elif not drop:
        _to_drop = set(columns) - set(always_keep)
    # elif isinstance(drop, str):
    #     _to_drop = [drop]
    elif isinstance(drop, Iterable) or isinstance(drop, list):
        _to_drop = drop
    else:
        raise Exception("Not valid subsetters!")

    _to_drop = set(_to_drop) - set(always_keep)

    # need to use unstrand to remove "+"/"-" from dict
    if "Strand" in _to_drop and self.stranded:
        self = self.unstrand()
        _to_drop = [d for d in _to_drop if not d == "Strand"]

    return self.apply(lambda df: df.drop(_to_drop, axis=1))


def _keep(self, keep):
    columns = self.columns
    if not self.stranded:
        always_keep = "Chromosome Start End".split()
    else:
        always_keep = "Chromosome Start End Strand".split()

    _to_drop = []
    if isinstance(keep, Iterable) or isinstance(keep, list):
        _to_drop = set(columns) - set(keep) - set(always_keep)
    else:
        raise Exception("Column(s) to keep must be in list.")

    self = self.apply(lambda df: df.drop(_to_drop, axis=1))

    columns = self.columns

    keep = [c for c in keep if not c in always_keep]
    new_order = []
    i = 0
    for c in columns:
        if c in always_keep:
            new_order.append(c)
        else:
            new_order.append(keep[i])
            i += 1

    self = self.apply(lambda df: df[new_order])

    return self



import numpy as np
import pandas as pd

from time import time

from sorted_nearest import (
    k_nearest_previous_nonoverlapping,
    k_nearest_next_nonoverlapping,
    get_all_ties,
)

try:
    dummy = profile
except NameError:
    profile = lambda x: x


def nearest_previous_idx(d1, d2, k, ties=None):
    d1s = d1.Start.sort_values()
    d2e = d2.End.sort_values()

    # ix where d1s could be inserted into d2e to preserve order of d2e
    # start = time()
    ix = np.searchsorted(d2e, d1s, side="right") - 1
    # end = time()
    # print("np previous", end - start)

    valid = ix >= 0
    ix = ix[valid]

    d1s = d1s.iloc[valid]

    lidx, ridx_pos, dist = k_nearest_previous_nonoverlapping(
        d1s.values, d2e.values, d1s.index.values, ix, k, ties
    )

    dist += 1

    ridx = d2e.iloc[ridx_pos].index

    return lidx, ridx, dist


# @profile
def nearest_next_idx(d1, d2, k, ties=None):
    d1e = d1.End.sort_values()
    d2s = d2.Start.sort_values()

    # ix where d1e could be inserted into d2s to preserve order of d2s
    # start = time()
    ix = np.searchsorted(d2s, d1e, side="left")
    # end = time()
    # print("np next", end - start)

    # - print(ix)
    valid = ix < len(d2s)
    ix = ix[valid]

    d1e = d1e.iloc[valid]

    lidx, ridx_pos, dist = k_nearest_next_nonoverlapping(
        d1e.values, d2s.values, d1e.index.values, ix, k, ties
    )

    dist += 1

    ridx = d2s.iloc[ridx_pos].index

    return lidx, ridx, dist


def nearest(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    plidx, pridx, pdist = nearest_previous_idx(d1, d2, d1.__k__.values, ties)
    nlidx, nridx, ndist = nearest_next_idx(d1, d2, d1.__k__.values, ties)

    pk = d1.__k__.reindex(plidx)
    nk = d1.__k__.reindex(nlidx)
    # px = d1.__IX__.reindex(plidx)
    # nx = d1.__IX__.reindex(nlidx)

    p = pd.DataFrame({"LX": plidx, "RX": pridx, "D": pdist, "PN": "P", "k": pk})
    n = pd.DataFrame({"LX": nlidx, "RX": nridx, "D": ndist, "PN": "N", "k": nk})

    df = pd.concat([p, n]).sort_values(["LX", "D"])
    # print(df)
    # raise

    if df.empty:
        return None

    xdf = df

    # row_indexer = xdf.PN == "P"
    # xdf.loc[row_indexer, "D"] = -xdf[row_indexer].D

    d1 = d1.reindex(xdf.LX)
    d2 = d2.reindex(xdf.RX)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    d2 = d2.drop("Chromosome", axis=1)
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", xdf.D.values)

    # to_drop = [c for c in df.columns if "__k__" in c]

    # df = df.drop(to_drop, axis=1)

    return df


def nearest_previous(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    lidx, ridx, dist = nearest_previous_idx(d1, d2, d1.__k__.values, ties)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d1.index = range(len(d1))
    d2.index = range(len(d1))

    d2 = d2.drop("Chromosome", axis=1)
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)

    # df = remove_duplicates_single(df, ties)

    # df.loc[:, "Distance"] = - df.Distance
    # to_drop = [c for c in df.columns if "__k__" in c]
    # df = df.drop(to_drop, axis=1)

    return df


def nearest_next(d1, d2, **kwargs):
    suffix = kwargs.get("suffix", "_b")
    ties = kwargs.get("ties", None)

    lidx, ridx, dist = nearest_next_idx(d1, d2, d1.__k__.values, ties)

    d1 = d1.reindex(lidx)
    d2 = d2.reindex(ridx)
    d2 = d2.drop("Chromosome", axis=1)
    d1.index = range(len(d1))
    d2.index = range(len(d1))
    df = d1.join(d2, rsuffix=suffix)
    df.insert(df.shape[1], "Distance", dist)

    return df


# @profile
def _nearest(d1, d2, **kwargs):
    if d1.empty or d2.empty:
        return None

    how = kwargs["how"]

    # if kwargs.get("strandedness") == "opposite":
    #     strand_dict = {"+": "-", "-": "+"}
    # else:
    #     strand_dict = {"+": "+", "-": "-"}

    if how in ["upstream", "downstream"] and kwargs["stranded"]:
        strand = d1.Strand.iloc[0]
        # strand = strand_dict[strand]

        __nearest = {
            ("+", "upstream"): nearest_previous,
            ("-", "upstream"): nearest_next,
            ("+", "downstream"): nearest_next,
            ("-", "downstream"): nearest_previous,
        }[strand, how]
    elif how in ["upstream", "downstream"] and not kwargs["stranded"]:
        __nearest = {"upstream": nearest_previous, "downstream": nearest_next}[how]
    else:
        __nearest = nearest

    df = __nearest(d1, d2, **kwargs)

    return df


if __name__ == "__main__":
    import pyranges as pr
    import numpy as np

    np.random.seed(0)
    chrM = pr.data.chromsizes()
    # chrM = chrM[chrM.Chromosome == "chrM"]
    size = int(1e5)
    print(np.log10(size))
    half_size = int(size / 2)
    strand = True

    gr = pr.random(size, chromsizes=chrM, strand=strand).sort()
    gr2 = pr.random(size, chromsizes=chrM, strand=strand).sort()
    gr.ID = np.arange(len(gr))
    gr2.ID = np.arange(len(gr2))

    from time import time

    start = time()
    ks = np.array([1, 2] * half_size, dtype=int)
    result = gr.k_nearest(gr2, k=ks, strandedness=None, overlap=True, ties="different")
    end = time()

    print(end - start)

    gr.msp()
    gr2.msp()

    result.msp()
    vc = result.ID.value_counts()
    # print(vc[vc > 2])
    # print(result[result.ID.isin(vc[vc > 2].index)])



import os
from distutils.core import setup
from setuptools import find_packages

__version__ = (
    open("pyranges/version.py").readline().split(" = ")[1].replace('"', "").strip()
)

install_requires = [
    "cython",
    "pandas",
    "ncls>=0.0.63",
    "tabulate",
    "sorted_nearest>=0.0.33",
    "pyrle",
    "natsort",
]  # ,

# optional_requires = ["bamread", "pybigwig", "ray"]

if os.getenv("TRAVIS"):
    install_requires.append("bamread pybigwig".split())

setup(
    name="pyranges",
    packages=find_packages(),
    package_data={
        "pyranges": [
            "example_data/*.bed",
            "example_data/*.gtf",
            "example_data/*.bam",
            "example_data/*.bam.bai",
            "example_data/*.gz",
            "example_data/*.bw",
        ]
    },
    # data_files = ["LICENSE"],
    include_dirs=["."],
    version=__version__,
    description="GenomicRanges for Python.",
    author="Endre Bakken Stovner",
    author_email="endrebak85@gmail.com",
    url="http://github.com/endrebak/pyranges",
    keywords=["Bioinformatics"],
    license="MIT",
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering",
    ],
    long_description=("Performant Pythonic GenomicRanges."),
)



import pandas as pd
import numpy as np

from ncls import NCLS


def add_rows_per_group(df):
    last_rows = df.groupby("__ix__").last().reset_index()
    last_rows.loc[:, "__last__"] = True
    df = pd.concat([df, last_rows], ignore_index=True)
    df = df.sort_values("__ix__", ascending=True)
    return df


# def _subtraction(scdf, **kwargs):
#     if scdf.empty:
#         return scdf

#     falses = np.zeros(len(scdf), dtype=bool)
#     scdf.insert(scdf.shape[1], "__first__", falses)
#     scdf.insert(scdf.shape[1], "__last__", falses)

#     scdf = add_rows_per_group(scdf)

#     scdf.insert(scdf.shape[1], "NewStart", scdf.End__deleteme__.shift(fill_value=-1))
#     scdf.insert(scdf.shape[1], "NewEnd", scdf.Start__deleteme__)
#     scdf.insert(scdf.shape[1], "__ix2__", np.arange(len(scdf)))

#     first_rows = scdf.groupby(scdf.__ix__, as_index=False).first()
#     scdf.loc[scdf.__ix2__.isin(first_rows.__ix2__), "__first__"] = True

#     scdf.loc[:, "NewStart"] = np.where(scdf.__first__, scdf.Start, scdf.NewStart)

#     scdf.loc[scdf.__first__ & ~(scdf.Start__deleteme__ >= scdf.Start), "NewStart"] = -1
#     scdf.loc[:, "NewEnd"] = np.where(scdf.__last__, scdf.End, scdf.NewEnd)
#     scdf.loc[:, "NewStart"] = np.where(scdf.__last__, scdf.End__deleteme__, scdf.NewStart)
#     scdf.loc[scdf.__last__ & ~(scdf.End__deleteme__ <= scdf.End), ["NewEnd", "NewStart"]] = -1


#     scdf = scdf[~((scdf.NewStart == -1) | (scdf.NewEnd == -1))]
#     scdf = scdf.drop(["Start", "End"], axis=1)
#     scdf.rename(columns={"NewStart": "Start", "NewEnd": "End"}, inplace=True)

#     remove_mask = scdf.Start >= scdf.End

#     return scdf[~remove_mask]


def _subtraction(scdf, ocdf, **kwargs):
    if ocdf.empty or scdf.empty:
        return scdf

    strandedness = kwargs["strandedness"]
    strand = True if strandedness else False

    chromosome = scdf.Chromosome.head(1).iloc[0]
    kwargs["chromosome"] = chromosome

    if "Strand" in ocdf and strand:
        strand = scdf.Strand.head(1).iloc[0]
        kwargs["strand"] = strand

    o = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    idx_self, new_starts, new_ends = o.set_difference_helper(
        scdf.Start.values, scdf.End.values, scdf.index.values, scdf.__num__.values
    )

    missing_idx = pd.Index(scdf.index).difference(idx_self)

    idx_to_drop = new_starts != -1

    new_starts = new_starts[idx_to_drop]
    new_ends = new_ends[idx_to_drop]

    idx_self = idx_self[idx_to_drop]
    new_starts = pd.Series(new_starts, index=idx_self)
    new_ends = pd.Series(new_ends, index=idx_self)

    scdf = scdf.reindex(missing_idx.union(idx_self)).sort_index()
    new_starts = new_starts.sort_index()
    new_ends = new_ends.sort_index()

    if len(idx_self):
        scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts.values
        scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends.values

    if not scdf.empty:
        return scdf
    else:
        return None



import pandas as pd

from sorted_nearest import find_clusters, merge_by


def _merge(df, **kwargs):
    if df.empty:
        return None

    slack = kwargs.get("slack", 0)
    chromosome, strand = kwargs["chromosome"], kwargs.get("strand", None)

    cdf = df.sort_values("Start")

    starts, ends, number = find_clusters(cdf.Start.values, cdf.End.values, slack)

    nidx = pd.Index(range(len(starts)))
    if strand:
        cluster_df = pd.DataFrame(
            {
                "Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
                "Start": starts,
                "End": ends,
                "Strand": pd.Series(strand, dtype="category", index=nidx),
            }
        )
    else:
        cluster_df = pd.DataFrame(
            {
                "Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
                "Start": starts,
                "End": ends,
            }
        )

    if kwargs["count"]:
        cluster_df.insert(cluster_df.shape[1], kwargs["count_col"], number)

    return cluster_df


def _merge_by(df, **kwargs):
    if df.empty:
        return None

    slack = kwargs.get("slack", 0)
    by = kwargs["by"]
    chromosome, strand = kwargs["chromosome"], kwargs.get("strand", None)

    if isinstance(by, str):
        cdf = df.sort_values([by, "Start"])
    else:
        cdf = df.sort_values(by + ["Start"])

    if isinstance(by, str):
        new_ids = (cdf[by] != cdf[by].shift()).cumsum()
    else:
        new_ids = (cdf[by] != cdf[by].shift()).any(axis=1).cumsum()

    new_to_old = pd.DataFrame(data={"new": new_ids.drop_duplicates()})

    new_to_old = pd.concat([new_to_old, cdf.reindex(new_to_old.index)[by]], axis=1)

    cdf.insert(cdf.shape[1], "ClusterBy", new_ids)

    ids, starts, ends, number = merge_by(
        cdf.Start.values, cdf.End.values, cdf.ClusterBy.values, slack
    )

    nidx = pd.Index(range(len(starts)))

    cluster_df = pd.DataFrame(
        {
            "Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
            "Start": starts,
            "End": ends,
            "by": ids,
        }
    )

    if strand:
        cluster_df.insert(
            cluster_df.shape[1],
            "Strand",
            pd.Series(strand, dtype="category", index=nidx),
        )

    cluster_df = cluster_df.merge(new_to_old, left_on="by", right_on="new")
    cluster_df = cluster_df.drop(["by", "new"], axis=1)
    cluster_df = cluster_df.rename(columns={"old": by})

    if kwargs["count"]:
        cluster_df.insert(cluster_df.shape[1], kwargs["count_col"], number)

    return cluster_df



import pyranges as pr
import pandas as pd

from collections import defaultdict


def concat(pyranges, strand=None):
    if not pyranges:
        return None

    pyranges = [pr for pr in pyranges if not pr.empty]
    grs_per_chromosome = defaultdict(list)

    strand_info = [gr.stranded for gr in pyranges]

    if strand is None:
        strand = all(strand_info)

    if strand:
        assert all(
            [gr.stranded for gr in pyranges]
        ), "Cannot do stranded concat, not all pyranges contain strand info."

        for gr in pyranges:
            for k, df in gr.dfs.items():
                # dbg(df)
                grs_per_chromosome[k].append(df)
    else:
        for gr in pyranges:
            for chromosome in gr.chromosomes:
                df = gr[chromosome].df
                grs_per_chromosome[chromosome].append(df)

    new_pyrange = {}

    for k, v in grs_per_chromosome.items():
        new_pyrange[k] = pd.concat(v, sort=False)

    res = pr.multithreaded.process_results(new_pyrange.values(), new_pyrange.keys())

    if any(strand_info) and not all(strand_info):
        new_res = {}
        for k, v in res.items():
            v.loc[:, "Strand"] = v.Strand.cat.add_categories(["."])
            new_res[k] = v.assign(Strand=v.Strand.fillna("."))
        res = pr.PyRanges(new_res)
        res.Strand = res.Strand
    else:
        res = pr.PyRanges(res)

    return res



from collections import OrderedDict

import pandas as pd

from tabulate import tabulate


def _summary(self, to_stdout=True, return_df=False):
    lengths = {}
    total_lengths = {}
    lengths["pyrange"] = self.lengths(as_dict=True)
    total_lengths["pyrange"] = [self.length]

    if self.stranded:
        c = self.merge(strand=True)
        lengths["coverage_forward"] = c["+"].lengths(as_dict=True)
        lengths["coverage_reverse"] = c["-"].lengths(as_dict=True)
        total_lengths["coverage_forward"] = [c["+"].length]
        total_lengths["coverage_reverse"] = [c["-"].length]
    else:
        c = self

    c = c.merge(strand=False)
    lengths["coverage_unstranded"] = c.lengths(as_dict=True)
    total_lengths["coverage_unstranded"] = [c.length]

    summaries = OrderedDict()

    # statistics for lengths
    for summary, d in lengths.items():
        if d:
            summaries[summary] = pd.concat(d.values()).describe()

    summary = pd.concat(summaries.values(), axis=1)
    summary.columns = list(summaries)

    df = pd.DataFrame.from_dict(total_lengths)
    df.index = ["sum"]
    summary = pd.concat([summary, df])

    if to_stdout:
        str_repr = tabulate(summary, headers=summary.columns, tablefmt="psql")
        print(str_repr)

    if return_df:
        return summary



from __future__ import print_function

import sys

import pandas as pd

from pyranges.pyranges_main import PyRanges

from io import StringIO

import pyranges as pr


def read_bed(f, as_df=False, nrows=None):
    """Return bed file as PyRanges.

    This is a reader for files that follow the bed format. They can have from
    3-12 columns which will be named like so:

    Chromosome Start End Name Score Strand ThickStart ThickEnd ItemRGB
    BlockCount BlockSizes BlockStarts

    Parameters
    ----------
    f : str

        Path to bed file

    as_df : bool, default False

        Whether to return as pandas DataFrame instead of PyRanges.

    nrows : int, default None

        Number of rows to return.

    Notes
    -----

    If you just want to create a PyRanges from a tab-delimited bed-like file,
    use `pr.PyRanges(pandas.read_table(f))` instead.

    Examples
    --------

    >>> path = pr.get_example_path("aorta.bed")
    >>> pr.read_bed(path, nrows=5)
    +--------------+-----------+-----------+------------+-----------+--------------+
    | Chromosome   |     Start |       End | Name       |     Score | Strand       |
    | (category)   |   (int32) |   (int32) | (object)   |   (int64) | (category)   |
    |--------------+-----------+-----------+------------+-----------+--------------|
    | chr1         |      9939 |     10138 | H3K27me3   |         7 | +            |
    | chr1         |      9953 |     10152 | H3K27me3   |         5 | +            |
    | chr1         |      9916 |     10115 | H3K27me3   |         5 | -            |
    | chr1         |      9951 |     10150 | H3K27me3   |         8 | -            |
    | chr1         |      9978 |     10177 | H3K27me3   |         7 | -            |
    +--------------+-----------+-----------+------------+-----------+--------------+
    Stranded PyRanges object has 5 rows and 6 columns from 1 chromosomes.
    For printing, the PyRanges was sorted on Chromosome and Strand.

    >>> pr.read_bed(path, as_df=True, nrows=5)
      Chromosome  Start    End      Name  Score Strand
    0       chr1   9916  10115  H3K27me3      5      -
    1       chr1   9939  10138  H3K27me3      7      +
    2       chr1   9951  10150  H3K27me3      8      -
    3       chr1   9953  10152  H3K27me3      5      +
    4       chr1   9978  10177  H3K27me3      7      -

    """

    columns = "Chromosome Start End Name Score Strand ThickStart ThickEnd ItemRGB BlockCount BlockSizes BlockStarts".split()

    if f.endswith(".gz"):
        import gzip

        first_start = gzip.open(f).readline().split()[1]
    else:
        first_start = open(f).readline().split()[1]

    header = None

    try:
        int(first_start)
    except ValueError:
        header = 0

    df = pd.read_csv(
        f,
        dtype={"Chromosome": "category", "Strand": "category"},
        nrows=nrows,
        header=header,
        sep="\t",
    )

    df.columns = columns[: df.shape[1]]

    if not as_df:
        return PyRanges(df)
    else:
        return df


def read_bam(f, sparse=True, as_df=False, mapq=0, required_flag=0, filter_flag=1540):
    """Return bam file as PyRanges.

    Parameters
    ----------
    f : str

        Path to bam file

    sparse : bool, default True

        Whether to return only.

    as_df : bool, default False

        Whether to return as pandas DataFrame instead of PyRanges.

    mapq : int, default 0

        Minimum mapping quality score.

    required_flag : int, default 0

        Flags which must be present for the interval to be read.

    filter_flag : int, default 1540

        Ignore reads with these flags. Default 1540, which means that either
        the read is unmapped, the read failed vendor or platfrom quality
        checks, or the read is a PCR or optical duplicate.

    Notes
    -----

    This functionality requires the library `bamread`. It can be installed with
    `pip install bamread` or `conda install -c bioconda bamread`.

    Examples
    --------

    >>> path = pr.get_example_path("control.bam")
    >>> pr.read_bam(path).sort()
    +--------------+-----------+-----------+--------------+------------+
    | Chromosome   | Start     | End       | Strand       | Flag       |
    | (category)   | (int32)   | (int32)   | (category)   | (uint16)   |
    |--------------+-----------+-----------+--------------+------------|
    | chr1         | 1041102   | 1041127   | +            | 0          |
    | chr1         | 2129359   | 2129384   | +            | 0          |
    | chr1         | 2239108   | 2239133   | +            | 0          |
    | chr1         | 2318805   | 2318830   | +            | 0          |
    | ...          | ...       | ...       | ...          | ...        |
    | chrY         | 10632456  | 10632481  | -            | 16         |
    | chrY         | 11918814  | 11918839  | -            | 16         |
    | chrY         | 11936866  | 11936891  | -            | 16         |
    | chrY         | 57402214  | 57402239  | -            | 16         |
    +--------------+-----------+-----------+--------------+------------+
    Stranded PyRanges object has 10,000 rows and 5 columns from 25 chromosomes.
    For printing, the PyRanges was sorted on Chromosome and Strand.

    >>> pr.read_bam(path, sparse=False, as_df=True)
         Chromosome    Start      End Strand  Flag  QueryStart  QueryEnd QuerySequence Name Cigar Quality
    0          chr1   887771   887796      -    16           0        25          None   U0   25M    None
    1          chr1   994660   994685      -    16           0        25          None   U0   25M    None
    2          chr1  1041102  1041127      +     0           0        25          None   U0   25M    None
    3          chr1  1770383  1770408      -    16           0        25          None   U0   25M    None
    4          chr1  1995141  1995166      -    16           0        25          None   U0   25M    None
    ...         ...      ...      ...    ...   ...         ...       ...           ...  ...   ...     ...
    9995       chrM     3654     3679      +     0           0        25          None   U0   25M    None
    9996       chrM     3900     3925      -    16           0        25          None   U0   25M    None
    9997       chrM    13006    13031      -    16           0        25          None   U0   25M    None
    9998       chrM    14257    14282      +     0           0        25          None   U0   25M    None
    9999       chrM    14257    14282      +     0           0        25          None   U0   25M    None
    <BLANKLINE>
    [10000 rows x 11 columns]
    """

    try:
        import bamread
    except ModuleNotFoundError as e:
        print(
            "bamread must be installed to read bam. Use `conda install -c bioconda bamread` or `pip install bamread` to install it."
        )
        sys.exit(1)

    if bamread.__version__ in [
        "0.0.1",
        "0.0.2",
        "0.0.3",
        "0.0.4",
        "0.0.5",
        "0.0.6",
        "0.0.7",
        "0.0.8",
        "0.0.9",
    ]:
        print(
            "bamread not recent enough. Must be 0.0.10 or higher. Use `conda install -c bioconda 'bamread>=0.0.10'` or `pip install bamread>=0.0.10` to install it."
        )
        sys.exit(1)

    if sparse:
        df = bamread.read_bam(f, mapq, required_flag, filter_flag)
    else:
        try:
            df = bamread.read_bam_full(f, mapq, required_flag, filter_flag)
        except AttributeError:
            print(
                "bamread version 0.0.6 or higher is required to read bam non-sparsely."
            )

    if as_df:
        return df
    else:
        return PyRanges(df)

    # return bamread.read_bam(f, mapq, required_flag, filter_flag)


def _fetch_gene_transcript_exon_id(attribute, annotation=None):
    no_quotes = attribute.str.replace('"', "").str.replace("'", "")

    df = no_quotes.str.extract(
        "gene_id.?(.+?);(?:.*transcript_id.?(.+?);)?(?:.*exon_number.?(.+?);)?(?:.*exon_id.?(.+?);)?",
        expand=True,
    )  # .iloc[:, [1, 2, 3]]

    df.columns = "gene_id transcript_id exon_number exon_id".split()

    if annotation == "ensembl":
        newdf = []
        for c in "gene_id transcript_id exon_id".split():
            r = df[c].astype(str).str.extract(r"(\d+)").astype(float)
            newdf.append(r)

        newdf = pd.concat(newdf, axis=1)
        newdf.insert(2, "exon_number", df["exon_number"])
        df = newdf

    return df


def skiprows(f):
    try:
        import gzip

        fh = gzip.open(f)
        for i, l in enumerate(fh):
            if l.decode()[0] != "#":
                break
    except (OSError, TypeError):  # not a gzipped file, or StringIO
        fh = open(f)
        for i, l in enumerate(fh):
            if l[0] != "#":
                break

    fh.close()

    return i


def read_gtf(f, full=True, as_df=False, nrows=None, duplicate_attr=False):
    """Read files in the Gene Transfer Format.

    Parameters
    ----------
    f : str

        Path to GTF file.

    full : bool, default True

        Whether to read and interpret the annotation column.

    as_df : bool, default False

        Whether to return as pandas DataFrame instead of PyRanges.

    nrows : int, default None

        Number of rows to read. Default None, i.e. all.

    duplicate_attr : bool, default False

        Whether to handle (potential) duplicate attributes or just keep last one.

    Note
    ----

    The GTF format encodes both Start and End as 1-based included.
    PyRanges (and also the DF returned by this function, if as_df=True), instead
    encodes intervals as 0-based, Start included and End excluded.

    See Also
    --------

    pyranges.read_gff3 : read files in the General Feature Format

    Examples
    --------

    >>> path = pr.get_example_path("ensembl.gtf")
    >>> gr = pr.read_gtf(path)

    >>> # +--------------+------------+--------------+-----------+-----------+------------+--------------+------------+-----------------+----------------+-------+
    >>> # | Chromosome   | Source     | Feature      | Start     | End       | Score      | Strand       | Frame      | gene_id         | gene_version   | +18   |
    >>> # | (category)   | (object)   | (category)   | (int32)   | (int32)   | (object)   | (category)   | (object)   | (object)        | (object)       | ...   |
    >>> # |--------------+------------+--------------+-----------+-----------+------------+--------------+------------+-----------------+----------------+-------|
    >>> # | 1            | havana     | gene         | 11868     | 14409     | .          | +            | .          | ENSG00000223972 | 5              | ...   |
    >>> # | 1            | havana     | transcript   | 11868     | 14409     | .          | +            | .          | ENSG00000223972 | 5              | ...   |
    >>> # | 1            | havana     | exon         | 11868     | 12227     | .          | +            | .          | ENSG00000223972 | 5              | ...   |
    >>> # | 1            | havana     | exon         | 12612     | 12721     | .          | +            | .          | ENSG00000223972 | 5              | ...   |
    >>> # | ...          | ...        | ...          | ...       | ...       | ...        | ...          | ...        | ...             | ...            | ...   |
    >>> # | 1            | ensembl    | transcript   | 120724    | 133723    | .          | -            | .          | ENSG00000238009 | 6              | ...   |
    >>> # | 1            | ensembl    | exon         | 133373    | 133723    | .          | -            | .          | ENSG00000238009 | 6              | ...   |
    >>> # | 1            | ensembl    | exon         | 129054    | 129223    | .          | -            | .          | ENSG00000238009 | 6              | ...   |
    >>> # | 1            | ensembl    | exon         | 120873    | 120932    | .          | -            | .          | ENSG00000238009 | 6              | ...   |
    >>> # +--------------+------------+--------------+-----------+-----------+------------+--------------+------------+-----------------+----------------+-------+
    >>> # Stranded PyRanges object has 95 rows and 28 columns from 1 chromosomes.
    >>> # For printing, the PyRanges was sorted on Chromosome and Strand.
    >>> # 18 hidden columns: gene_name, gene_source, gene_biotype, transcript_id, transcript_version, transcript_name, transcript_source, transcript_biotype, tag, transcript_support_level, ... (+ 8 more.)
    """

    _skiprows = skiprows(f)

    if full:
        gr = read_gtf_full(f, as_df, nrows, _skiprows, duplicate_attr)
    else:
        gr = read_gtf_restricted(f, _skiprows, as_df=False, nrows=None)

    return gr


def read_gtf_full(f, as_df=False, nrows=None, skiprows=0, duplicate_attr=False):
    dtypes = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    names = "Chromosome Source Feature Start End Score Strand Frame Attribute".split()

    df_iter = pd.read_csv(
        f,
        sep="\t",
        header=None,
        names=names,
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=skiprows,
        nrows=nrows,
    )

    _to_rows = to_rows_keep_duplicates if duplicate_attr else to_rows

    dfs = []
    for df in df_iter:
        extra = _to_rows(df.Attribute)
        df = df.drop("Attribute", axis=1)
        ndf = pd.concat([df, extra], axis=1, sort=False)
        dfs.append(ndf)

    df = pd.concat(dfs, sort=False)
    df.loc[:, "Start"] = df.Start - 1

    if not as_df:
        return PyRanges(df)
    else:
        return df


def to_rows(anno):
    rowdicts = []
    try:
        l = anno.head(1)
        for l in l:
            l.replace('"', "").replace(";", "").split()
    except AttributeError:
        raise Exception(
            "Invalid attribute string: {l}. If the file is in GFF3 format, use pr.read_gff3 instead.".format(
                l=l
            )
        )

    for l in anno:
        rowdicts.append(
            {
                k: v
                for k, v in [
                    kv.replace('""', '"NA"').replace('"', "").split(None, 1)
                    # rstrip: allows for GFF not having a last ";", or having final spaces
                    for kv in l.rstrip("; ").split("; ")
                ]
            }
        )

    return pd.DataFrame.from_dict(rowdicts).set_index(anno.index)


def to_rows_keep_duplicates(anno):
    rowdicts = []
    for l in anno:
        rowdict = {}

        # rstrip: allows for GFF not having a last ";", or having final spaces
        for k, v in (
            kv.replace('"', "").split(None, 1) for kv in l.rstrip("; ").split("; ")
        ):
            if k not in rowdict:
                rowdict[k] = v
            elif k in rowdict and isinstance(rowdict[k], list):
                rowdict[k].append(v)
            else:
                rowdict[k] = [rowdict[k], v]

        rowdicts.append(
            {k: ",".join(v) if isinstance(v, list) else v for k, v in rowdict.items()}
        )

    return pd.DataFrame.from_dict(rowdicts).set_index(anno.index)


def read_gtf_restricted(f, skiprows, as_df=False, nrows=None):
    """seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    # source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    # frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature.
    """
    dtypes = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    df_iter = pd.read_csv(
        f,
        sep="\t",
        comment="#",
        usecols=[0, 2, 3, 4, 5, 6, 8],
        header=None,
        names="Chromosome Feature Start End Score Strand Attribute".split(),
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=skiprows,
        nrows=nrows,
    )

    dfs = []
    for df in df_iter:
        if sum(df.Score == ".") == len(df):
            cols_to_concat = "Chromosome Start End Strand Feature".split()
        else:
            cols_to_concat = "Chromosome Start End Strand Feature Score".split()

        extract = _fetch_gene_transcript_exon_id(df.Attribute)
        extract.columns = "gene_id transcript_id exon_number exon_id".split()

        extract.exon_number = extract.exon_number.astype(float)

        df = pd.concat([df[cols_to_concat], extract], axis=1, sort=False)

        dfs.append(df)

    df = pd.concat(dfs, sort=False)

    df.loc[:, "Start"] = df.Start - 1

    if not as_df:
        return PyRanges(df)
    else:
        return df


def to_rows_gff3(anno):
    rowdicts = []

    for l in list(anno):
        # stripping last white char if present
        lx = (it.split("=") for it in l.rstrip("; ").split(";"))
        rowdicts.append({k: v for k, v in lx})

    return pd.DataFrame.from_dict(rowdicts).set_index(anno.index)


def read_gff3(f, full=True, annotation=None, as_df=False, nrows=None):
    """Read files in the General Feature Format.

    Parameters
    ----------
    f : str

        Path to GFF file.

    full : bool, default True

        Whether to read and interpret the annotation column.

    as_df : bool, default False

        Whether to return as pandas DataFrame instead of PyRanges.

    nrows : int, default None

        Number of rows to read. Default None, i.e. all.

    Notes
    -----

    The gff3 format encodes both Start and End as 1-based included.
    PyRanges (and also the DF returned by this function, if as_df=True), instead
    encodes intervals as 0-based, Start included and End excluded.

    See Also
    --------

    pyranges.read_gtf : read files in the Gene Transfer Format
    """

    _skiprows = skiprows(f)

    if not full:
        return read_gtf_restricted(f, _skiprows, as_df=as_df, nrows=nrows)

    dtypes = {"Chromosome": "category", "Feature": "category", "Strand": "category"}

    names = "Chromosome Source Feature Start End Score Strand Frame Attribute".split()

    df_iter = pd.read_csv(
        f,
        comment="#",
        sep="\t",
        header=None,
        names=names,
        dtype=dtypes,
        chunksize=int(1e5),
        skiprows=_skiprows,
        nrows=nrows,
    )

    dfs = []
    for df in df_iter:
        extra = to_rows_gff3(df.Attribute.astype(str))
        df = df.drop("Attribute", axis=1)
        ndf = pd.concat([df, extra], axis=1, sort=False)
        dfs.append(ndf)

    df = pd.concat(dfs, sort=False)

    df.loc[:, "Start"] = df.Start - 1

    if not as_df:
        return PyRanges(df)
    else:
        return df


def read_bigwig(f, as_df=False):
    try:
        import pyBigWig
    except ModuleNotFoundError:
        print(
            "bwread must be installed to read bigwigs. Use `conda install -c bioconda bwread` or `pip install bwread` to install it."
        )
        import sys

        sys.exit(1)

    """Read bigwig files.

    Parameters
    ----------
    f : str

        Path to bw file.

    as_df : bool, default False

        Whether to return as pandas DataFrame instead of PyRanges.

    Examples
    --------

    >>> f = pr.get_example_path("bw.bw")
    >>> gr = pr.read_bigwig(f)
    >>> gr
    """

    from natsort import natsorted

    bw = pyBigWig.open(f)

    size = int(1e5)
    chromosomes = bw.chroms()

    dfs = {}

    for chromosome in natsorted(chromosomes):
        outstarts = []
        outends = []
        outvalues = []

        length = chromosomes[chromosome]

        starts = list(range(0, length, size))
        ends = list(range(size, length + size, size))
        ends[-1] = length
        for start, end in zip(starts, ends):
            intervals = bw.intervals(chromosome, start, end)
            if intervals is not None:
                for s, e, v in intervals:
                    outstarts.append(s)
                    outends.append(e)
                    outvalues.append(v)

        outstarts = pd.Series(outstarts)
        outends = pd.Series(outends)
        outvalues = pd.Series(outvalues)
        dfs[chromosome] = pd.DataFrame(
            {
                "Chromosome": chromosome,
                "Start": outstarts,
                "End": outends,
                "Value": outvalues,
            }
        )

    return pr.PyRanges(dfs)



import pyranges as pr


def test_mcc():
    gr1, gr2 = pr.data.chipseq(), pr.data.chipseq_background()
    g = pr.data.chromsizes()
    mcc = pr.stats.mcc([gr1, gr2], genome=g, labels=["chip", "bg"], strand=True)

    print(mcc)
